import os
import click
import time
import signal
import re
from util import Config
from util import Device
from util import Util
import zeroconf as Zeroconf

devices_found = []

# discovers all devices exposing stak on local network and
# returns information about them
def discover():

  # callback to add devices to a list when stak devices are found
  def zc_service_change(zeroconf, service_type, name, state_change):
    if state_change is Zeroconf.ServiceStateChange.Added:
      info = zeroconf.get_service_info(service_type, name)
      dev_type = "generic"
      if "rpi" in name:
        dev_type = "Raspberry Pi"
      devices_found.append( { "name": name, "info": info, "type": dev_type } )
    elif state_change is Zeroconf.ServiceStateChange.Removed:
      pass

  # create a zeroconf object
  zeroconf = Zeroconf.Zeroconf()

  # stup a zeroconf browser to search for _stak._tcp.local.
  browser = Zeroconf.ServiceBrowser(zeroconf, "_stak._tcp.local.", handlers=[zc_service_change] )
  click.echo("Searching for stak-enabled devices on local network...")

  # show a progress bar for 5 seconds while we search
  with click.progressbar( length=40, width=10, show_percent=True,
                          show_eta=False,
                          fill_char=click.style('#', fg='magenta')) as bar:
    for item in bar:
      time.sleep(0.125)

  # end zeroconf communications
  zeroconf.close()

  # if we found any devices
  if len(devices_found) == 0:
    click.echo( click.style("Error: No stak devices found!", bold=True, fg="red" ) )
    return

  Util.display_list( "Devices found:", devices_found,
    ( lambda index,value:
      click.style( value["info"].server, bold=True, fg="cyan" )
      + ": "
      + click.style( value["type"], bold=True, fg="red" )
    )
  )

  # ask the user to select one
  new_device = devices_found[ click.prompt("Please select the device to add", type=int) ]
  click.echo("Selected: " + new_device["name"])

  #
  new_device["name"] = re.sub('\._stak\._tcp\.local.$', '', new_device["name"])

  #
  uuid = new_device["info"].properties["uuid"] if ( "uuid" in new_device["info"].properties ) else ""

  #
  add ( new_device["name"], uuid, new_device["info"].server, "root")


# add a device to the project configuration
# 
def add( device, uuid=None, address=None, user=None, key_file=None):

  # if the device already exists in the configuration
  # then we don't want to add it again.
  # otherwise, prep data for adding device
  if not Device.exists( Config.local_data, device ):

    set_as_default = False
    if Device.count( Config.local_data ) == 0:
      set_as_default = True

    # create base device data structure
    device_data = {}

    # if any of the data we specified was None,
    # we want to set the data to sane defaults
    device_data['uuid'] = uuid if uuid else str( Util.call_and_return("uuidgen") ).rstrip()
    device_data['address'] = address if address else "stak.local"
    device_data['user'] = user if user else "root"
    device_data['key-file'] = key_file if key_file else os.path.expanduser("~/.ssh/id_rsa.pub")
    
    # add device to local data
    Config.local_data = Device.add( Config.local_data, device, device_data )

    if set_as_default:
      set_default_device( device )

    # save local project config data/Users/zerotri/Development/ntc/stak-cli/util/Device.py
    Config.save( )

def set_default_device( device ):
  if not Device.exists( Config.local_data, device ):
    click.echo("Device "+device+" doesn't exist!")
    return
  if not Config.local_data:
    click.echo("Config data not being set properly!")
    return
  Config.local_data["default-device"] = device
  Config.save( )

def set_property( device, parameter, value ):
  if Device.exists( Config.local_data, device ):
    Device.set( Config.local_data, device, parameter, value )
    Config.save( )

def get_property( device, parameter ):
  if Device.exists( Config.local_data, device ):
    prop = Device.get( Config.local_data, device, parameter )
    if prop:
      click.echo( prop )
    else:
      click.echo( "Property doesn't exist!" )

def remove( device ):
  if Device.exists( Config.local_data, device ):
    Config.local_data = Device.remove( Config.local_data, device )
    Config.save( )
        
#
# device list
#
def list( ):
  if Config.verbose:
    print("Verbose")

  if len( Config.local_data['devices'] ) < 1:
    click.echo( "No devices found" )
    return

  Util.display_list( "Listing devices:", Config.local_data['devices'],
    (
      lambda index,value: click.style( value, bold=True, fg="cyan" )
        + ": "
        + click.style( Device.get( Config.local_data, value, 'uuid' ), fg="green" )
    )
  )

# finds valid sd cards inserted in the current system
# currently OSX only as it relies on diskutil
# TODO@Wynter: Add support for Linux and Windows
def find_valid_sd_cards():
  sd_cards = []
  diskutil_output = Util.call_and_return("diskutil", "list")
  disks = Util.find_list_items_containing( "^\/dev\/disk[0-9]$", diskutil_output )
  for disk in disks:
    diskutil_output = Util.call_and_return("diskutil", "info", disk)
    if Util.find_list_items_containing("(Protocol:\s+Secure Digital)", diskutil_output):
      sd_cards.append( disk )
    if Util.find_list_items_containing("(Protocol:\s+USB)", diskutil_output):
      sd_cards.append( disk )
  return sd_cards

# finds all fat32 partitions inserted in the system
# you may optionally pass a disk name in to have it
# only find fat partitions on the specified disk
def find_fat_partitions(disk=None):
  valid_paths = []
  diskutil_args = ["diskutil", "list"]
  if disk:
    diskutil_args.append( disk )

  diskutil_output = Util.call_and_return( *diskutil_args )
  search_regex = "\d\:\s+.*FAT_32\s+(.+)\s+\d+\.\d+\s+[kKgGmM]B+.*(disk[0-9]s[0-9])"
  fat_partitions = Util.find_list_items_containing( search_regex,
    diskutil_output,
    ( lambda m: [m.group(1).strip(), m.group(2)] )
  )
  return fat_partitions

# flash the specified disk with the stak os image.
def flash_card( disk ):

  # we need to unmount all partitions on the specified disk
  # or the os will complain about the disks being in use
  Util.call_and_return( "diskutil","unmountDisk", disk )

  # find the location of the raspbian image.
  # this is located inside of a hidden folder in the user directory for now
  # TODO@Wynter: we should also verify that the file exists and if not download it
  raspbian_image = os.path.expanduser("~/.stak/images/netinst-image.iso")

  # output some information for the user to see that something is going on
  # TODO@Wynter: get some sort of progress output from DD
  click.echo("Writing...")

  # flash the disk
  rdisk = disk.replace("/dev/disk", "/dev/rdisk")

  if rdisk:
    Util.dd( raspbian_image, rdisk )
  else:
    Util.dd( raspbian_image, disk )

def get_latest_os_image():
  http_header = Util.call_and_return( "curl", "-I", "http://stak-images.s3.amazonaws.com/netinst-image.iso.bz2")
  etag = Util.find_list_items_containing("^ETag:\s\"(.*)\"", http_header,
    ( lambda m: m.group(1) ))
  if not etag:
    click.echo( "Nothing in etag" )
    return
  
  image_path = os.path.expanduser("~/.stak/images/netinst-image.iso.bz2")
  click.echo( Config.global_data["latest-checksum"] )
  if( "latest-checksum" in Config.global_data and
      Config.global_data["latest-checksum"] == etag[0] and
      os.path.exists( os.path.expanduser("~/.stak/images/netinst-image.iso") ) ):
    click.echo("Using the latest image. Continuing...")
  else:
    click.echo("Downloading image...")
    Util.dd( "http://stak-images.s3.amazonaws.com/netinst-image.iso.bz2", image_path )

    click.echo("Decompressing image...")
    Util.call_and_return("bunzip2", image_path )
    Config.global_data["latest-checksum"] = etag[0]
    Config.save_globals()

# locate and set up an sd card with a new stak os image
def setup_card():

  if os.getuid() != 0:
    click.echo("Not running as super user. Please relaunch as super user")
    return

  inserted=False

  cards = None
  click.wrap_text("text\x08", width=78, initial_indent='', subsequent_indent='', preserve_paragraphs=True)
  click.echo(click.style('WARNNING!! This command will delete EVERYTHING on the target volume', reverse=True))

  # find and verify that the card is inserted
  while not inserted:
    if not click.confirm('Is the SD card inserted?'):
      click.echo("Please insert the SD card now.")
      click.pause()
    else:
      cards = find_valid_sd_cards()
      if not cards:
        click.echo("No sd card found!")
        click.echo("Please insert the SD card now.")
        click.pause()
      else:
        inserted = True

  # have user select the disk to use
  Util.display_list( "Cards found", cards )
  selected_card = cards[ click.prompt("WARNNING!! Make SURE this is the correct card, this command will delete EVERYTHING on the target volume. Please select the card to use", type=int) ]
  click.echo("Selected: " + selected_card)

  get_latest_os_image()

  # flash the selected sd card with the stak os base image
  flash_card( selected_card )

  # sleep for five seconds to make sure the os has time to
  # register and mount the new partitions on the volume
  time.sleep(20.0)

  # find fat partitions on the selected disk
  fat_partitions = find_fat_partitions( selected_card )
  if not fat_partitions:
    click.echo("No partitions found! Card likely corrupt.")
    return

  # begin copying extra files over to new volume
  volume = "/Volumes/" + fat_partitions[0][0]
  click.echo("Writing to card: " + volume)

  # copy the post-install.txt and installer-config.txt used by
  # raspbian-ua-netinst
  #post_install = os.path.expanduser("~/.stak/post-install.txt")
  #install_config = os.path.expanduser("~/.stak/installer-config.txt")
  
  #Util.cp( post_install, volume+"/post-install.txt" )
  #Util.cp( install_config, volume+"/installer-config.txt" )

  # copy the current user's public key to the sd card
  key_file = os.path.expanduser("~/.ssh/id_rsa.pub")
  
  # if volume:
  if not os.path.exists(key_file):
    Util.call_and_return( "ssh-keygen","-t", "rsa" "-f", key_file )
  click.echo("Writing key file: " + key_file)
  Util.cp( key_file, volume+"/id_rsa.pub" )
  if not os.path.exists( volume+"/id_rsa.pub" ):
    click.echo("ERROR COPYING KEY!")
  click.echo("Copied key!")



import click
import yaml
import os
import re
import subprocess
import DD

# displays a list of data.
# you can control how the data is output by 
def display_list( string, list , transform = ( lambda index,value: click.style( value, bold=True, fg="cyan" ) ) ):
  # list devices
    click.echo( click.style( string , bold=True) )
    for index, value in enumerate( list ):
      click.echo( "  " + str(index) + ") " + transform( index, value ) )

# calls a shell command and returns the output
def call_and_return( *args, **kwargs ):
  if "directory" in kwargs:
    return subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=kwargs["directory"]).stdout.readlines()
  else:
    return subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.readlines()


# utility function for searching a list for a regex.
# you can also pass a lambda in to control how the data
# is transformed before being added to the resulting list
def find_list_items_containing( needle, haystack, transform=( lambda m: m.group() ) ):
  output = []
  for item in haystack:
    matches = re.search( needle, item )
    if matches:
      output.append( transform( matches ) )
  if len(output) == 0:
    return None
  return output


def cp( src, dest ):
  call_and_return( "cp",src, dest )

def create_folders_if_nonexistent( *folders ):
  for folder in folders:
    folder_expanded = os.path.expanduser( folder )
    if not os.path.exists( folder_expanded ):
      os.mkdir( folder_expanded )


def dd( src, dst ):
  # defaults
  iput  = DD.FileIo('/dev/stdin', 'r')
  oput  = DD.FileIo('/dev/stdout', 'a')
  ts    = None
  bs    = 2048000 # 4000KB
  count = None
  skip  = None
  quiet = False

  if os.path.isfile(src):
    iput = DD.FileIo(src, 'r')
  else:
    if src.find('http'):
      iput = DD.HttpIo( src )
    elif src.find('ftp'):
      iput = DD.FtpIo ( src )

  oput = DD.FileIo(dst, 'w+')
  
  obj = DD.Transfer(iput, oput, ts, bs, count, skip, quiet)
  obj.start()



import click
import yaml
import os

global_config_file = os.path.expanduser( "~/.stak/config.yml" )
example_languages = [ 'c', 'c++', 'python', 'javascript' ]
global_data = {}
local_data = {}
verbose = False
stak_directory = click.Path('~')

def load_globals():
  global global_data, global_config_file
  if os.path.isfile( global_config_file ):
    global_file = click.open_file( global_config_file, "r+", lazy=True, encoding = "utf-8")
    if global_file:
      config_data = yaml.safe_load( global_file.read() )
      global_data = config_data
    #   if config.data is None:
    #     config.data = {}  
    #   if config.data.get('devices', None) is None:
    #     config.data["devices"] = {}
  else:
    global_file = click.open_file("stak.yml", "w", lazy=True, encoding = "utf-8")
    global_data = {}

def save_globals( ):
  global global_data, global_config_file
  config_file = click.open_file( global_config_file, "w", lazy=True, encoding = "utf-8" )
  yaml.safe_dump( global_data, config_file, default_flow_style=False, encoding='utf-8', allow_unicode=True )

def save( directory = None, config_data = None ):
  global local_data
  if not directory:
    directory = os.getcwd()
  if not config_data:
    config_data = local_data

  local_file = click.open_file( directory + "/stak.yml", "w", lazy=True, encoding = "utf-8")
  yaml.safe_dump( config_data, local_file, default_flow_style=False, encoding='utf-8', allow_unicode=True )

def load( directory = None ):
  if not directory:
    directory = os.getcwd()

  if os.path.isfile( directory + "/stak.yml" ):
    local_file = click.open_file( directory + "/stak.yml", "r+", lazy=True, encoding = "utf-8")
    data = None
    if local_file:
      data = yaml.safe_load( local_file.read() )
      config_data = data

    data["devices"] = data["devices"] if ("devices" in config_data and config_data["devices"]) else {}
    config_data = data if (data) else {}
  else:
    config_data = {}
    config_data["devices"] = {}
  return config_data

def get_attributes( attributes, config = None):
  global local_data
  if not config:
    config = local_data

  return_attr = {}

  for attr in attributes:
    if attr[0] in config:
      return_attr[ attr[0] ] = config[attr[0] ]
    else:
      click.echo( attr[0]+" not defined in project config" )
      return_attr[ attr[0] ] = attr[1]
  return return_attr


import os
import click
import time
import sys
from util import Config
from util import Project
from util import Util
from distutils import spawn
from commands import *

#
# Stak command group
#
@click.group()
@click.option('--verbose', is_flag=True , help="Prints extra information for each command.")
@click.option('--stak-directory', type=click.Path(), help="Location of your .stak directory.")
def stak( verbose, stak_directory):
  """STAK is a collection of tools for building projects on embedded Linux"""

  Config.verbose = verbose
  Config.stak_directory = stak_directory
  Config.local_data = Config.load( )
  Config.load_globals( )
  Config.current_device = None

  # make sure we have the following directories
  # ~/.stak
  # ~/.stak/images
  Util.create_folders_if_nonexistent( "~/.stak", "~/.stak/images" )


  # test if honcho exists
  if not spawn.find_executable("honcho"):
    click.echo("Could not find honcho executable. Exiting...")
    sys.exit()


#def iterate():
#  number = 0
#  while number < 1000:
#    yield number
#    number+=1
#
#def show_iteration(item):
#  if item:
#    return "Iteration: %d" % item
#
#@stak.command()
#def jump():
#  with click.progressbar(iterable=iterate(), width=10, show_percent=False,
#                          show_eta=False, item_show_func=show_iteration,
#                          fill_char=click.style('#', fg='magenta')) as bar:
#    for item in bar:
#      time.sleep(0.0313)

#
# Create command
#
@stak.command()
@click.argument('project')
@click.option('--language',
              default=None,
              type=click.Choice( Config.example_languages )
  )
def create( project, language ):
  """Creates a new STAK project."""
  Create.create( project, language )

#
# device [command]
#
@stak.group()
def device():
  """Tools for managing hardware running STAK OS"""
  pass


@device.command("setup-card")
def setup_card():
  """Downloads and flashes the STAK OS"""
  Devices.setup_card()
#
# device add [device]
#
@device.command()
@click.argument('device')
def add( device ):
  Devices.add( device )
#
# device remove [device]
#
@device.command()
@click.argument('device')
def remove( device ):
  """Remove device from project"""
  Devices.remove( device )

@device.command("set-default")
@click.argument('device')
def set_default( device ):
  Devices.set_default_device( device )
#
#
# device list
#
@device.command()
def list( ):
  """List all STAK devices on your network"""
  Devices.list( )

#
# device discover
#
#   search the network for any devices broadcasting a stak zeroconf service
#
@device.command()
def discover( ):
  Devices.discover()


@device.command()
@click.argument('name', default=None, required=False)
@click.argument('parameter', type=click.STRING)
@click.argument('value', type=click.STRING)
def set( name, parameter, value ):
  if name:
    Devices.set_property(name, parameter, value)

@device.command()
@click.argument('name', default=None, required=False)
@click.argument('parameter', type=click.STRING)
def get( name, parameter ):
  if name:
    Devices.get_property(name, parameter)


#
# Build command
#
@stak.command()
@click.argument('project',
                required=False,
                default=os.getcwd(),
                type=click.Path()
  )
@click.option('--target',
                required=False,
                default="debug",
                type=click.STRING
  )
def build( project, target ):
  """Cross-compile your project"""
  Project.build( project, target )
  # click.echo("Comming Soon to Terminal Near You...")
#
# Run command
#
@stak.command()
@click.argument('project',
                required=False,
                default=os.getcwd(),
                type=click.Path()
  )
@click.option('--target',
                required=False,
                default="debug",
                type=click.STRING
  )
@click.option('--device',type=click.STRING)
def run( project, target, device ):
  """Run code built with STAK BUILD on device"""
  Project.run( project, target )

#
# ssh command
#
@stak.command()
@click.argument('project',
                required=False,
                default=os.getcwd(),
                type=click.Path()
  )
@click.option('--target',
                required=False,
                default="debug",
                type=click.STRING
  )
@click.option('--device',type=click.STRING)
def ssh( project, target, device ):
  """Open SSH session on device"""
  Project.ssh( project, target )


#
# Deploy command
#
@stak.command()
@click.argument('project',
                required=False,
                default=os.getcwd(),
                type=click.Path()
  )
@click.option('--target',
                required=False,
                default="debug",
                type=click.STRING
  )
@click.option('--device',type=click.STRING)
def deploy( project, target, device ):
  """Installs project as system service that runs on BOOT"""
  Project.deploy( project, target )


import click
import os
from util import Config
from util import Project

#
# create
#   should 
#
def create( project, language ):
  # set directory variable
  directory = project

  # setup example options
  example_options = {}

  # make sure directory doesn't already exist
  if os.path.exists(directory):
    click.echo( "Cannot create directory " + directory
              + " because it already exists")
    return

  # prompt to create an example project
  example_options['create'] = click.confirm("Would you like an example project created for you?")
  if example_options['create']:
    if language is not None:
      example_options['language'] = language
    else:
      click.echo( click.style("Supported languages:", bold=True) )
      for index, lang in enumerate( Config.example_languages ):
        click.echo( "  "
          + str(index)
          + ") "
          + click.style(lang, bold=True, fg="cyan")
        )
      example_options['language'] = Config.example_languages[click.prompt("Please select the language to use", type=int)]
      click.echo("Selected: " + example_options['language'])

  # create directory
  Project.create( project, example_options )


from setuptools import setup, setuptools

setup(
    name="stak",
    version="0.1.0a2",
    py_modules=['stak'],
    install_requires=["Click","PyYAML","zeroconf","jinja2<2.8,>=2.7", "honcho", "honcho[export]"],
    description='stak cli tool',
    long_description="Command line interface for the stak development platform",
    url='http://nextthing.co/stak.html',
    author='Next Thing Co.',
    author_email='Ahoyahoy@nextthing.co',
    license='MIT',
    packages=setuptools.find_packages(exclude=['lib', 'include', 'bin','Stak.egg-info']),
    keywords='sample setuptools development',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
         'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    entry_points='''
    [console_scripts]
    stak=stak:stak
    '''
)